from typing import List, Tuple, Optional, Union
import logging
import numpy as np
from PIL import Image
from pymilvus import Collection

# 引入特征提取器和MilvusClient
from feature_extractor import FeatureExtractor
from milvus import MilvusClient # 假设MilvusClient在milvus_client.py文件中

class ImageSearch:
    """
    图像搜索类，提供基于图像相似度的搜索功能
    """

    _logger = logging.getLogger(__name__)
    _feature_extractor_initialized = False
    _milvus_client = None

    @classmethod
    def _initialize_feature_extractor(cls) -> None:
        """初始化特征提取器"""
        if not cls._feature_extractor_initialized:
            try:
                FeatureExtractor.initialize()
                cls._feature_extractor_initialized = True
            except Exception as e:
                cls._logger.error(f"Failed to initialize feature extractor: {e}")
                raise RuntimeError(f"Failed to initialize feature extractor: {e}")

    @classmethod
    def get_milvus_client(cls, host: str = "localhost", port: str = "19530",
                          collection_name: str = "image_collection") -> MilvusClient:
        """
        获取或创建Milvus客户端

        Args:
            host: Milvus服务器地址
            port: Milvus服务器端口
            collection_name: 集合名称

        Returns:
            MilvusClient: Milvus客户端实例
        """
        if cls._milvus_client is None:
            cls._milvus_client = MilvusClient(host, port, collection_name).connect()
        return cls._milvus_client

    @classmethod
    def get_collection(cls, host: str = "localhost", port: str = "19530",
                       collection_name: str = "image_collection") -> Collection:
        """
        获取Milvus集合

        Args:
            host: Milvus服务器地址
            port: Milvus服务器端口
            collection_name: 集合名称

        Returns:
            Collection: Milvus集合对象

        Raises:
            RuntimeError: 如果无法获取集合
        """
        try:
            client = cls.get_milvus_client(host, port, collection_name)
            return client.get_collection()
        except Exception as e:
            cls._logger.error(f"Failed to get collection: {e}")
            raise RuntimeError(f"Failed to get collection: {e}")

    @classmethod
    def extract_features(cls, image: Union[str, Image.Image]) -> np.ndarray:
        """
        从图像提取特征向量

        Args:
            image: 图像URL或PIL图像对象

        Returns:
            np.ndarray: 特征向量

        Raises:
            ValueError: 如果特征提取失败
        """
        # 确保特征提取器已初始化
        cls._initialize_feature_extractor()

        try:
            # 根据图像类型调用相应的提取方法
            if isinstance(image, str):
                features = FeatureExtractor.extract_from_url(image)
            elif isinstance(image, Image.Image):
                features = FeatureExtractor.extract_from_image(image)
            else:
                raise ValueError(f"Unsupported image type: {type(image)}")

            if features is None:
                raise ValueError("Feature extraction returned None")

            return features

        except Exception as e:
            cls._logger.error(f"Feature extraction failed: {e}")
            raise ValueError(f"Failed to extract features: {e}")

    @classmethod
    def image_to_image_search(
            cls,
            image: Union[str, Image.Image],
            top_k: int = 100,
            host: str = "localhost",
            port: str = "19530",
            collection_name: str = "image_collection"
    ) -> List[Tuple[str, float]]:
        """
        使用图像查询相似图像

        Args:
            image: 查询图像的URL或PIL图像对象
            top_k: 返回的最相似结果数量
            host: Milvus服务器地址
            port: Milvus服务器端口
            collection_name: 集合名称

        Returns:
            List[Tuple[str, float]]: 产品ID和相似度分数的列表，按相似度降序排列

        Raises:
            ValueError: 如果图像处理或搜索过程中出错
        """
        try:
            # 提取查询图像的特征
            query_embedding = cls.extract_features(image)

            # 获取Milvus客户端并搜索
            client = cls.get_milvus_client(host, port, collection_name)
            results = client.search(query_embedding, limit=top_k)

            # 处理结果
            if not results or len(results) == 0:
                return []

            # 返回结果
            product_ids = [hit.entity.get('product_id') for hit in results[0]]
            scores = [hit.score for hit in results[0]]

            return list(zip(product_ids, scores))

        except Exception as e:
            cls._logger.error(f"Image search failed: {e}")
            raise ValueError(f"Image search failed: {e}")

    @classmethod
    def batch_image_search(
            cls,
            images: List[Union[str, Image.Image]],
            top_k: int = 100,
            host: str = "localhost",
            port: str = "19530",
            collection_name: str = "image_collection"
    ) -> List[List[Tuple[str, float]]]:
        """
        批量图像搜索

        Args:
            images: 查询图像URL或PIL图像对象的列表
            top_k: 每个查询返回的最相似结果数量
            host: Milvus服务器地址
            port: Milvus服务器端口
            collection_name: 集合名称

        Returns:
            List[List[Tuple[str, float]]]: 每个查询图像对应的结果列表
        """
        # 确保特征提取器已初始化
        cls._initialize_feature_extractor()

        # 批量提取特征
        batch_embeddings = []
        valid_indices = []

        for i, image in enumerate(images):
            try:
                features = cls.extract_features(image)
                batch_embeddings.append(features)
                valid_indices.append(i)
            except Exception as e:
                cls._logger.warning(f"Feature extraction failed for image at index {i}: {e}")

        # 如果没有有效的特征，返回空列表
        if not batch_embeddings:
            return [[] for _ in range(len(images))]

        try:
            # 获取集合
            collection = cls.get_collection(host, port, collection_name)
            collection.load()

            # 批量搜索
            search_params = {"metric_type": "IP", "params": {"ef": 100}}
            batch_results = collection.search(
                data=batch_embeddings,
                anns_field="embedding",
                param=search_params,
                limit=top_k,
                output_fields=["image", "product_id"]
            )

            # 处理结果
            all_results = [[] for _ in range(len(images))]

            for i, results in enumerate(batch_results):
                original_idx = valid_indices[i]
                product_ids = [hit.entity.get('product_id') for hit in results]
                scores = [hit.score for hit in results]
                all_results[original_idx] = list(zip(product_ids, scores))

            return all_results

        except Exception as e:
            cls._logger.error(f"Batch image search failed: {e}")
            raise ValueError(f"Batch image search failed: {e}")
