from typing import Dict, List, Any, Optional, Union
import numpy as np
from pymilvus import connections, Collection, FieldSchema, CollectionSchema, DataType, utility


class MilvusClient:
    def __init__(self, host: str = "localhost", port: str = "19530", collection_name: str = "image_collection") -> None:
        """初始化Milvus客户端

        Args:
            host: Milvus服务器地址
            port: Milvus服务器端口
            collection_name: 集合名称
        """
        self.host: str = host
        self.port: str = port
        self.collection_name: str = collection_name
        self.collection: Optional[Collection] = None

    def connect(self) -> 'MilvusClient':
        """连接到Milvus服务器

        Returns:
            MilvusClient: 当前客户端实例，支持链式调用
        """
        connections.connect("default", host=self.host, port=self.port)
        return self

    def get_collection(self) -> Collection:
        """获取或创建集合

        Returns:
            Collection: Milvus集合对象
        """
        # 定义集合结构
        fields: List[FieldSchema] = [
            FieldSchema(name="image", dtype=DataType.VARCHAR, max_length=256, is_primary=True, auto_id=False),
            FieldSchema(name="product_id", dtype=DataType.VARCHAR, max_length=256),
            FieldSchema(name="embedding", dtype=DataType.FLOAT_VECTOR, dim=512)  # CLIP ViT-B/32的特征维度为512
        ]
        schema: CollectionSchema = CollectionSchema(fields, "图像特征集合")

        # 创建或获取集合
        if utility.has_collection(self.collection_name):
            self.collection = Collection(name=self.collection_name)
        else:
            self.collection = Collection(name=self.collection_name, schema=schema)
            self._create_index()

        return self.collection

    def _create_index(self) -> None:
        """创建索引"""
        index_params: Dict[str, Any] = {
            "index_type": "HNSW",
            "metric_type": "IP",  # 内积相似度
            "params": {"M": 16, "efConstruction": 200}
        }
        if self.collection is not None:
            self.collection.create_index(field_name="embedding", index_params=index_params)

    def search(self, vector: Union[List[float], np.ndarray], limit: int = 10) -> Any:
        """搜索相似向量

        Args:
            vector: 查询向量，可以是列表或numpy数组
            limit: 返回结果数量

        Returns:
            查询结果
        """
        if self.collection is None:
            self.get_collection()

        self.collection.load()
        search_params: Dict[str, Any] = {"metric_type": "IP", "params": {"ef": 100}}
        results = self.collection.search(
            data=[vector],
            anns_field="embedding",
            param=search_params,
            limit=limit,
            output_fields=["image", "product_id"]
        )
        return results

    def insert(self, data: List[Dict[str, Any]]) -> None:
        """插入数据

        Args:
            data: 包含image、product_id和embedding的字典列表
                 每个字典应包含键: "image", "product_id", "embedding"
        """
        if self.collection is None:
            self.get_collection()

        entities: List[List[Any]] = [
            [item["image"] for item in data],
            [item["product_id"] for item in data],
            [item["embedding"] for item in data]
        ]

        self.collection.insert(entities)

    def drop_collection(self) -> None:
        """删除集合"""
        if utility.has_collection(self.collection_name):
            utility.drop_collection(self.collection_name)
            self.collection = None

    def close(self) -> None:
        """关闭连接"""
        connections.disconnect("default")
